import typing

from abqpy.decorators import abaqus_class_doc, abaqus_method_doc
from ..UtilityAndView.abaqusConstants import *


@abaqus_class_doc
class SurfaceCrushTriggerAssignment:
    """The SurfaceCrushTriggerAssignment object stores the surface crush trigger assignment
    definition for surfaces in ContactExp objects. The SurfaceCrushTriggerAssignment object
    has no constructor or members.

    .. note::
        This object can be accessed by::

            import interaction
            mdb.models[name].interactions[name].surfaceCrushTriggerAssignments

        The corresponding analysis keywords are:

        - SURFACE PROPERTY ASSIGNMENT

    .. versionadded:: 2021
        The `SurfaceCrushTriggerAssignment` class was added.
    """

    @abaqus_method_doc
    def changeValuesInStep(
        self, stepName: str, index: int, value: typing.Union[SymbolicConstant, float]
    ):
        """This method allows modification of surface crush trigger assignments already defined on
        surfaces in a given step.

        Parameters
        ----------
        stepName
            A String specifying the name of the step in which the surface crush trigger assignments
            are to be modified.
        index
            An Int specifying the position of the surface crush trigger assignment whose value is to
            be modified.
        value
            A tuple specifying the value of the surface crush trigger assignments for the surface
            whose index is referenced. Each tuple contains three entries:
            - A SymbolicConstant specifying the trigger option to be used for the surface. Possible
            values of the SymbolicConstant are TRIGGER, NO_TRIGGER, or NO_CRUSH.
            - A Float specifying the crush stress value to be used for the surface.
            - A Float specifying the crush initiation angle value to be used for the surface.
            - A Float specifying the crush continuation angle value to be used for the surface.
        """
        ...

    @abaqus_method_doc
    def appendInStep(
        self, stepName: str, assignments: typing.Union[SymbolicConstant, float]
    ):
        """This method allows addition of surface crush trigger assignments to new surfaces in a
        given step.

        Parameters
        ----------
        stepName
            A String specifying the name of the step in which new surface crush trigger assignments
            are to be defined.
        assignments
            A sequence of tuples specifying the surface crush trigger assignments. Each tuple
            contains four entries:
            - A region or a material object or the SymbolicConstant GLOBAL specifying the surface to
            which the feature angle is assigned.
            - A SymbolicConstant specifying the trigger option to be used for the surface. Possible
            values of the SymbolicConstant are TRIGGER, NO_TRIGGER, or NO_CRUSH.
            - A Float specifying the crush stress value to be used for the surface.
            - A Float specifying the crush initiation angle value to be used for the surface.
            - A Float specifying the crush continuation angle value to be used for the surface.
        """
        ...

    @abaqus_method_doc
    def delete(self, indices: tuple):
        """The delete method allows you to delete existing surface crush trigger assignments from a
        ContactExp object.

        Parameters
        ----------
        indices
            A sequence of Ints specifying the index of each surface crush trigger assignment to
            delete.
        """
        ...



import json
import os
from typing import Dict, Union, Type
import textwrap

import toml
from jinja2 import Template

__all__ = [
    'CompressionTemplate',
    'ScriptTemplate',
]


def load_json_or_toml(path: Union[str, Dict]) -> Dict:
    """Load a json or toml file."""
    if isinstance(path, dict):
        return path
    if not os.path.exists(path):
        raise FileNotFoundError(f'File {path} does not exist.')
    if path.endswith('.json'):
        with open(path, encoding='utf-8') as f:
            return json.load(f)
    elif path.endswith('.toml'):
        with open(path, encoding='utf-8') as f:
            return toml.load(f)
    else:
        raise ValueError(f'Unknown file type: {path}')


class ScriptTemplate(Template):
    """A template class for Abaqus script.

    Examples
    --------
    >>> from abqpy.template import ScriptTemplate
    >>> template = ScriptTemplate(template='template.py', config='config.toml', params='params.toml')
    >>> template.write('script.py')
    >>> template = ScriptTemplate(template='template.py', config='config.toml')
    >>> template.write('script.py', param1=1, param2=2)
    """
    #: Parameter config
    _config: Dict[str, Dict[str, Union[str, int, float, bool]]]
    #: User parameters config
    _params: Dict[str, Union[str, int, float, bool]]
    #: Template source
    _template_source: str
    #: Parameter config source
    _config_source: str
    #: User config source
    _params_source: str
    #: A list of parameters required by the template.
    keys = property(lambda self: list(self._config))
    #: The default parameters.
    defaults = property(lambda self: {name: param['default'] for name, param in self._config.items()})
    #: The types of the parameters.
    types = property(lambda self: {name: param['type'] for name, param in self._config.items()})
    #: The descriptions of the parameters.
    descriptions = property(lambda self: {name: param['description'] for name, param in self._config.items()})
    #: Template source
    template_source = property(lambda self: self._template_source)
    #: Parameter config source
    config_source = property(lambda self: self._config_source)
    #: User config source
    params_source = property(lambda self: self._params_source)

    def __new__(
        cls,
        template: Union[
            os.PathLike, str,
        ],
        config: Union[
            Dict[str, Dict[str, Union[str, int, float, bool]]],
            os.PathLike, str,
        ] = None,
        params: Union[
            Dict[str, Dict[str, Union[str, int, float, bool]]],
            os.PathLike, str,
        ] = None,
    ):
        """Create a new template.

        Parameters
        ----------
        template : str or PathLike or Template
            The template source.
        config : str or PathLike or dict, optional
            The config file (a json or toml file) or dict, by default None
        params : str or PathLike or dict, optional
            The user params config file (a json or toml file) or dict, by default None
        """
        if os.path.exists(template) and os.path.isfile(config):
            template = open(template, 'r', encoding='utf-8').read()
        obj = super().__new__(cls, template)
        obj._template_source = template

        # Read the config file
        if os.path.exists(config) and os.path.isfile(config):
            obj._config_source = open(config, 'r', encoding='utf-8').read()
        obj._config = load_json_or_toml(config) if config else {}
        for name in obj._config.keys():
            obj._config[name]['name'] = name

        # Read the user config file
        if os.path.exists(params) and os.path.isfile(params):
            obj._params_source = open(params, 'r', encoding='utf-8').read()
        obj._params = load_json_or_toml(params) if params else {}
        return obj

    def check(self, correct_bounds=True, **kwargs):
        """Check if the parameters are valid, if valid, return the processed parameters as a dict.

        Parameters
        ----------
        correct_bounds : bool, optional
            Whether to correct the parameters to the bounds, by default True
        **kwargs
            The parameters to check.

        Returns
        -------
        Dict
            The original or corrected parameters.

        Raises
        ------
        ValueError
            If the parameters are invalid.
        """
        for name, param in kwargs.items():
            if name not in self._config:
                raise ValueError(f'Invalid parameter: {name}')
            if 'min' in self._config[name] and param < self._config[name]['min']:
                if correct_bounds:
                    kwargs[name] = self._config[name]['min']
                else:
                    raise ValueError(f'Parameter {name} is too small: {param}')
            if 'max' in self._config[name] and param > self._config[name]['max']:
                if correct_bounds:
                    kwargs[name] = self._config[name]['max']
                else:
                    raise ValueError(f'Parameter {name} is too large: {param}')
        params = self.defaults
        params.update(self._params)
        params.update(kwargs)
        return params

    def render(self, **kwargs) -> str:
        """Render the template.

        Parameters
        ----------
        **kwargs
            The parameters to render the template.

        Returns
        -------
        str
            The rendered template.
        """
        return super().render(**self.check(**kwargs))

    def write(self, file: Union[os.PathLike, str], **kwargs):
        """Write the rendered template to a file.

        Parameters
        ----------
        file : str or PathLike
            The file to write to.
        **kwargs
            The parameters to render the template.
        """
        with open(file, 'w', encoding='utf-8') as f:
            f.write(self.render(**kwargs))


def template_doc(cls: Type['CompressionTemplate']):
    """Generate the docstring for the template class."""
    obj = cls()
    attr_docstrings = []
    keys, types, defaults, descriptions = obj.keys, obj.types, obj.defaults, obj.descriptions
    for key in keys:
        attr_docstrings.append(f'.. confval:: {key}\n'
                               f'    :type: {types[key]}, defaults to {defaults[key]}\n\n'
                               f'    {descriptions[key]}')
    attrs_docstring = '\n\n'.join(attr_docstrings)
    docstring = f"""
This is a template for {cls.name}. 

Examples
--------
>>> from abqpy.template import {cls.__name__}
>>> template = {cls.__name__}(params='params.toml')
>>> template.write('script.py')
>>> template = {cls.__name__}()
>>> template.write('script.py', param1=1, param2=2)

.. admonition:: Details of required parameters

{textwrap.indent(attrs_docstring, ' ' * 4)}

.. admonition:: User config file (A `toml <https://toml.io/>`_ or `json <https://www.json.org/>`_ file)
    
    .. code-block:: toml
    
{textwrap.indent(obj.params_source, ' ' * 8)}

.. admonition:: The template source code

    .. code-block:: python
    
{textwrap.indent(obj.template_source, ' ' * 8)}

.. admonition:: The template config file  (A `toml <https://toml.io/>`_ or `json <https://www.json.org/>`_ file)
    
    .. code-block:: toml
    
{textwrap.indent(obj.config_source, ' ' * 8)}
"""
    cls.__doc__ = docstring
    return cls


class _DocumentTemplate(ScriptTemplate):
    #: The name of the template, the corresponding template filename must be `name.tmpl`, and the corresponding config
    #: filename must be `name.toml`.
    name = 'template'

    def __new__(
        cls,
        params: Union[
            Dict[str, Dict[str, Union[str, int, float, bool]]],
            os.PathLike, str,
        ] = None,
    ):
        dirname = os.path.dirname(__file__)
        template = os.path.join(dirname, 'templates', f'{cls.name}.tmpl')
        config = os.path.join(dirname, 'templates', f'{cls.name}.conf.toml')
        if params is None:
            params = os.path.join(dirname, 'templates', f'{cls.name}.toml')
        return super().__new__(cls, template, config, params)


@template_doc
class CompressionTemplate(_DocumentTemplate):
    name = 'compression'


def test_render():
    """Test the render function."""
    os.chdir(os.path.dirname(__file__))
    os.chdir('../../')
    template = ScriptTemplate('templates/compression.tmpl', 'templates/compression.toml')
    print(template.render(width=2, length=2, height=2))


def test_decorator():
    """Test the decorator."""
    print(CompressionTemplate.__doc__)



from abqpy.decorators import abaqus_class_doc


@abaqus_class_doc
class RandomResponseFrequency:
    """A RandomResponseFrequency is an object used to define frequency over a range of modes.
    This page discusses:

    .. note:: 
        This object can be accessed by::

            import step
            mdb.models[name].steps[name].freq[i]
    """

    #: A Float specifying the lower limit of the frequency range in cycles per time.
    lower: float = None

    #: A Float specifying the upper limit of the frequency range in cycles per time.
    upper: float = None

    #: An Int specifying the number of points between eigenfrequencies at which the response
    #: should be calculated.
    nCalcs: int = None

    #: A Float specifying the bias parameter.
    bias: float = None



from abqpy.decorators import abaqus_class_doc, abaqus_method_doc

@abaqus_class_doc
class MeshFace:
    """The MeshFace object refers to an element face. It has no constructor or members. A
    MeshFace object can be accessed via a MeshFaceArray or a repository on a part or part
    instance.

    .. note:: 
        This object can be accessed by::

            import part
            mdb.models[name].parts[name].elementFaces[i]
            mdb.models[name].parts[name].elemFaces[i]
            import assembly
            mdb.models[name].rootAssembly.allInstances[name].elementFaces[i]
            mdb.models[name].rootAssembly.allInstances[name].elemFaces[i]
            mdb.models[name].rootAssembly.instances[name].elementFaces[i]
            mdb.models[name].rootAssembly.instances[name].elemFaces[i]
    """

    #: An Int specifying an Int specifying the element label.
    label: int = None

    #: An Int specifying a symbolic constant specifying the side of the element.
    face: int = None

    @abaqus_method_doc
    def getElemEdges(self):
        """This method returns a tuple of unique element edges on the element face.

        Returns
        -------
        edges: typing.Tuple[MeshEdge, ...]
            A tuple of MeshEdge objects
        """
        ...

    @abaqus_method_doc
    def getElements(self):
        """This method returns a tuple of elements that share the element face.

        Returns
        -------
        elements: typing.Tuple[MeshElement, ...]
            A tuple of MeshElement objects
        """
        ...

    @abaqus_method_doc
    def getNodes(self):
        """This method returns a tuple of nodes on the element face.

        Returns
        -------
        nodes: typing.Tuple[MeshNode, ...]
            A tuple of MeshNode objects
        """
        ...

    @abaqus_method_doc
    def getNodesByFaceAngle(self, angle: str):
        """This method returns an array of mesh node objects that are obtained by recursively
        finding adjacent element faces that are at an angle of less than or equal to the
        specified angle.

        Parameters
        ----------
        angle
            A float specifying the value of the face angle.

        Returns
        -------
        nodes: MeshNodeArray
            A :py:class:`~abaqus.Mesh.MeshNodeArray.MeshNodeArray` object, which is a sequence of MeshNode objects
        """
        ...

    @abaqus_method_doc
    def getNormal(self):
        """This method returns the normal direction for the element face.

        Returns
        -------
        normal: typing.Tuple[float, ...]
            A tuple of 3 floats representing the unit normal vector. If the element face is
            collapsed such that a normal cannot be computed, a zero-length vector is returned.
        """
        ...

    @abaqus_method_doc
    def getElemFacesByFaceAngle(self, angle: str):
        """This method returns an array of element face objects that are obtained by recursively
        finding adjacent element faces that are at an angle of less than or equal to the
        specified angle.

        Parameters
        ----------
        angle
            A float specifying the value of the face angle.

        Returns
        -------
        faces: MeshFaceArray
            A :py:class:`~abaqus.Mesh.MeshFaceArray.MeshFaceArray` object, which is a sequence of MeshFace objects.
        """
        ...

    @abaqus_method_doc
    def getElemEdgesByFaceAngle(self, angle: str):
        """This method returns an array of element edge objects that are obtained by recursively
        finding adjacent element edges that are at an angle of less than or equal to the
        specified face angle.

        Parameters
        ----------
        angle
            A float specifying the value of the face angle in degrees.

        Returns
        -------
        edges: MeshEdgeArray
            A :py:class:`~abaqus.Mesh.MeshEdgeArray.MeshEdgeArray` object, which is a sequence of MeshEdge objects.
        """
        ...

    @abaqus_method_doc
    def getElementsByFaceAngle(self, angle: str):
        """This method returns an array of mesh Element objects that are obtained by recursively
        finding adjacent element faces that are at an angle of less than or equal to the
        specified angle.

        Parameters
        ----------
        angle
            A float specifying the value of the face angle.

        Returns
        -------
        elements: MeshElementArray
            A :py:class:`~abaqus.Mesh.MeshElementArray.MeshElementArray` object, which is a sequence of MeshElement objects.
        """
        ...

    @abaqus_method_doc
    def getElemFacesByLimitingAngle(self, angle: str):
        """This method returns an array of element edge objects that are obtained by recursively
        finding adjacent element faces that are at an angle of less than or equal to the
        specified face angle with the seed face.

        Parameters
        ----------
        angle
            A float specifying the value of the face angle in degrees.

        Returns
        -------
        faces: MeshFaceArray
            A :py:class:`~abaqus.Mesh.MeshFaceArray.MeshFaceArray` object, which is a sequence of MeshFace objects.
        """
        ...

    @abaqus_method_doc
    def getElementsViaTopology(self):
        """This method returns an array of mesh Element objects that are obtained by recursively
        finding adjacent elements via topology.

        Returns
        -------
        elements: MeshElementArray
            A :py:class:`~abaqus.Mesh.MeshElementArray.MeshElementArray` object, which is a sequence of MeshElement objects.
        """
        ...

    @abaqus_method_doc
    def getElemFacesByLayer(self, numLayers: str):
        """This method returns an array of element face objects, obtained by traversing shell
        elements or the exterior of a solid mesh, and recursively finding adjacent element faces
        by layer.

        Parameters
        ----------
        numLayers
            A int specifying the value of the number of layers.

        Returns
        -------
        faces: MeshFaceArray
            A :py:class:`~abaqus.Mesh.MeshFaceArray.MeshFaceArray` object, which is a sequence of MeshFace objects.
        """
        ...



from abqpy.decorators import abaqus_class_doc
from .BoundaryConditionState import BoundaryConditionState
from ..UtilityAndView.abaqusConstants import *


@abaqus_class_doc
class TemperatureBCState(BoundaryConditionState):
    """The TemperatureBCState object stores the propagating data for a temperature boundary
    condition in a step. One instance of this object is created internally by the
    TemperatureBC object for each step. The instance is also deleted internally by the
    TemperatureBC object.
    The TemperatureBCState object has no constructor or methods.
    The TemperatureBCState object is derived from the BoundaryConditionState object.

    .. note:: 
        This object can be accessed by::

            import load
            mdb.models[name].steps[name].boundaryConditionStates[name]

        The corresponding analysis keywords are:

        - BOUNDARY
    """

    #: A Float specifying the temperature magnitude.
    magnitude: float = None

    #: A SymbolicConstant specifying the propagation state of the temperature magnitude.
    #: Possible values are UNSET, SET, UNCHANGED, FREED, and MODIFIED.
    magnitudeState: SymbolicConstant = None

    #: A SymbolicConstant specifying the propagation state of the **dof** member. Possible values
    #: are SET and UNCHANGED.
    dofState: SymbolicConstant = None

    #: A tuple of Ints specifying the degrees of freedom to which the boundary condition is
    #: applied.
    dof: int = None

    #: A SymbolicConstant specifying the propagation state of the amplitude reference. Possible
    #: values are UNSET, SET, UNCHANGED, FREED, and MODIFIED.
    amplitudeState: SymbolicConstant = None

    #: A SymbolicConstant specifying the propagation state of the BoundaryConditionState object. Possible values are:
    #: 
    #: - NOT_YET_ACTIVE
    #: - CREATED
    #: - PROPAGATED
    #: - MODIFIED
    #: - DEACTIVATED
    #: - NO_LONGER_ACTIVE
    #: - TYPE_NOT_APPLICABLE
    #: - INSTANCE_NOT_APPLICABLE
    #: - PROPAGATED_FROM_BASE_STATE
    #: - MODIFIED_FROM_BASE_STATE
    #: - DEACTIVATED_FROM_BASE_STATE
    #: - BUILT_INTO_MODES
    status: SymbolicConstant = None

    #: A String specifying the name of the amplitude reference. The String is empty if the
    #: boundary condition has no amplitude reference.
    amplitude: str = ""



from abqpy.decorators import abaqus_class_doc


@abaqus_class_doc
class EmagTimeHarmonicFrequency:
    """

    .. note:: 
        This object can be accessed by::

            import step
            mdb.models[name].steps[name].frequencyRange[i]
    """

    #: A Float specifying the lower limit of frequency range or a single frequency, in
    #: cycles/time.
    lower: float = None

    #: A Float specifying the upper limit of frequency range, in cycles/time.
    upper: float = None

    #: An Int specifying the number of points in the frequency range at which results should be
    #: given.
    nPoints: int = None



import os
import sys


def run(cae: bool = True) -> None:
    """Runs Abaqus command in system's CLI
    
    This function uses the top level script file to run the Abaqus
    command, and the arguments passed to it
    
    Parameters
    ----------
    cae : bool, optional
        Whether to use `abaqus cae` command or
        `abaqus python`, by default True
    """
    abaqus = os.environ.get("ABAQUS_BAT_PATH", "abaqus")
    try:
        filePath = os.path.abspath(sys.modules['__main__'].__file__)
    except (AttributeError, TypeError, Exception):
        return 
    args = " ".join(sys.argv[1:])

    try:  # If it is a jupyter notebook
        import ipynbname
        filePath = ipynbname.path()
        os.system(f"jupyter nbconvert --to python {filePath}")
        filePath = str(filePath).replace(".ipynb", ".py")
    except (FileNotFoundError, ImportError, Exception):
        pass

    # check if in debug mode and run
    debug = os.environ.get("ABQPY_DEBUG", "false").lower() == "true"
    gettrace = getattr(sys, 'gettrace', None)
    if not debug and gettrace is not None and gettrace():
        debug = True

    # Check if it is imported by sphinx to generate docs
    make_docs = os.environ.get("ABQPY_MAKE_DOCS", "false").lower() == "true"

    # If in debug mode do not run the abaqus command at all
    if not debug and not make_docs:
        if cae:
            os.system(f"{abaqus} cae noGUI={filePath} -- {args}")
        else:
            os.system(f"{abaqus} python {filePath} {args}")
        sys.exit(0)



from abqpy.decorators import abaqus_class_doc, abaqus_method_doc
from .Leaf import Leaf
from ..UtilityAndView.abaqusConstants import *


@abaqus_class_doc
class LeafFromOdbElementTypes(Leaf):
    """The LeafFromOdbElementTypes object can be used whenever a Leaf object is expected as an
    argument. Leaf objects are used to specify the items in a display group. Leaf objects
    are constructed as temporary objects, which are then used as arguments to DisplayGroup
    commands.
    The LeafFromOdbElementTypes object is derived from the Leaf object.

    .. note:: 
        This object can be accessed by::

            import displayGroupOdbToolset
    """

    #: A SymbolicConstant specifying the leaf type. Possible values are EMPTY_LEAF,
    #: DEFAULT_MODEL, ALL_ELEMENTS, ALL_NODES, and ALL_SURFACES.
    leafType: SymbolicConstant = None

    @abaqus_method_doc
    def __init__(self, elementTypes: tuple):
        """This method creates a Leaf object from a sequence of Strings specifying element names.
        Leaf objects specify the items in a display group.

        .. note:: 
            This function can be accessed by::

                LeafFromOdbElementTypes

        Parameters
        ----------
        elementTypes
            A sequence of Strings specifying element Types.

        Returns
        -------
        LeafFromOdbElementTypes
            A :py:class:`~abaqus.DisplayGroup.LeafFromOdbElementTypes.LeafFromOdbElementTypes` object.
        """
        super().__init__(DEFAULT_MODEL)
